/*
 * Copyright (c) 2018, Dropbox, Inc. All rights reserved.
 */

package com.dropbox.android.camerauploads

import android.content.Context
import com.dropbox.android.camerauploads.CameraUploadsManager.DEFAULT_BATTERY_THRESHOLD_VALUE
import com.dropbox.android.notifications.SystemTrayNotifications
import com.dropbox.android.provider.SDKProvider
import com.dropbox.android.settings.GlobalProperties
import com.dropbox.android.user.DbxUser
import com.dropbox.android.util.RuntimePermissionPrompter
import com.dropbox.base.analytics.CameraUploadsFeatureEvents
import com.dropbox.base.oxygen.DbxLog
import com.dropbox.core.legacy_api.exception.DropboxException

internal abstract class CameraUploadsCoreControllerBase(
    protected val context: Context,
    protected val user: DbxUser,
    protected val jobManager: CameraUploadJobManager
) : CameraUploadsCoreControllerFacade {
    /**
     * A helper method to enable camera uploads, first prompting for storage permissions if needed.
     * @param runtimePermissionsHelper for requesting the WRITE_PHOTOS permission
     * @param uploadExisting whether CU should upload the existing photos
     * @param use3g whether CU should use 3G
     * @param uploadVideos whether CU should upload videos
     * @param successCallback an optional callback when CU was enabled successfully. Will be called
     *                        with a boolean parameter which is true if CU wasn't enabled previously
     *                        and the user setting was changed.
     */
    protected fun enableCameraUploadsForUserIfPermissionGranted(
        runtimePermissionsHelper: RuntimePermissionPrompter,
        uploadExisting: Boolean,
        use3g: Boolean,
        uploadVideos: Boolean,
        successCallback: Runnable?
    ) {
        // Ensure the user has storage permissions
        runtimePermissionsHelper
                .process(
                        object : RuntimePermissionPrompter.Callback {
                            override fun permissionGranted() {
                                enableCameraUploadsForUser(
                                        uploadExisting,
                                        use3g,
                                        uploadVideos)

                                successCallback?.run()
                            }

                            override fun permissionDenied() {}
                        })
    }

    /**
     * Disables camera uploads for the given user, if it is not already disabled.
     * @param turnedOffByApp Indicates whether CU was turned off by the app (true) or whether the
     *                      user chose to disable (false)
     * @return Whether CU was previously enabled and the CU setting was changed
     */
    protected fun disableCameraUploadsForUserIfEnabled(turnedOffByApp: Boolean): Boolean {
        val didUpdate = updateCameraUploadEnabledSettingForUser(enabled = false)

        if (didUpdate) {
            jobManager.setAllJobsFinished()
            jobManager.stop()
            val trayNotificationController = user.systemTrayNotificationController
            // Clear any camera upload notifications since we're turning off.
            trayNotificationController.clear(
                    user.id, SystemTrayNotifications.Singular.CAMERA_UPLOAD_STOPPED_QUOTA)
            trayNotificationController.clear(
                    user.id, SystemTrayNotifications.Singular.CAMERA_UPLOAD_PAUSED_BATTERY)
            trayNotificationController.clear(
                    user.id, SystemTrayNotifications.Singular.CAMERA_UPLOAD_PAUSED_UNPLUGGED)

            user.logger.log(CameraUploadsFeatureEvents.Disabled())

            SDKProvider.notifyCameraUploadSettingsChange(context)

            if (turnedOffByApp) {
                // TODO(ageiduschek): Migrate this from SharedProperties to UserProperties
                val sharedProps = user.getIdentityState().getSharedProperties()
                sharedProps.setShouldShowCameraUploadTurnedOffModal(true)
                // Show the notification
                trayNotificationController
                        .show(user.id,
                                SystemTrayNotifications.Singular.CAMERA_UPLOAD_TURN_OFF_NOTICE,
                                null)
                user.logger.log(CameraUploadsFeatureEvents.TurnedOffByApp())
            }
        }

        return didUpdate
    }

    private fun enableCameraUploadsForUser(
        uploadExisting: Boolean,
        use3g: Boolean,
        uploadVideos: Boolean
    ) {
        val userProps = user.userProperties

        userProps.setCameraUploadUse3G(use3g)
        userProps.setCameraUploadVideosEnabled(uploadVideos)
        // Send the new setting to the server asynchronously.
        asyncSendVideoUploadEnabledSetting(user, uploadVideos)

        userProps.setCameraUploadBatteryThreshold(DEFAULT_BATTERY_THRESHOLD_VALUE.toLong())
        userProps.setCameraUploadLimitVideoUploadsOverCell(true)

        // START legacy CU user settings
        // TODO(ageiduschek) Remove these once we roll CU V3 out to 100%
        userProps.setCameraUploadFirstMediaScan(true)
        userProps.setCameraUploadInitialScan(true)
        userProps.setCameraUploadIgnoreExisting(!uploadExisting)
        userProps.setCameraUploadHashUpdate(GlobalProperties.HashUpdate.OPTIONAL)
        userProps.setCameraUploadNotification(
                if (uploadExisting) GlobalProperties.CUNotification.AFTER_BACKLOG
                else GlobalProperties.CUNotification.AFTER_FIRST_SEVERAL)
        userProps.resetCameraUploadNumUploads()
        userProps.setCameraUploadHadBacklog(false)
        // END legacy CU user settings

        val didUpdate = updateCameraUploadEnabledSettingForUser(enabled = true)

        if (didUpdate) {
            jobManager.start(user)
            jobManager.enqueueJob(user, CameraUploadJobManager.EnqueueJobReason.SETTING_ENABLED)

            user.logger.log(CameraUploadsFeatureEvents.Enabled())

            SDKProvider.notifyCameraUploadSettingsChange(context)
        }
    }

    /**
     * @return True if the user exists and the setting changed, false otherwise.
     */
    private fun updateCameraUploadEnabledSettingForUser(
        enabled: Boolean
    ): Boolean {
        val userProperties = user.userProperties
        // Do nothing if there is no change
        if (userProperties.cameraUploadEnabled() == enabled) {
            return false
        }

        userProperties.setCameraUploadEnabled(enabled)

        return true
    }

    /**
     * Make a best-effort attempt to send the new video enabled setting to the server to be used for
     * future defaults. The api call happens in another thread, so this is safe to call from any thread
     * and returns immediately without checking for success. There is no retrying in case of failure.
     */
    fun asyncSendVideoUploadEnabledSetting(user: DbxUser, videosEnabled: Boolean) {
        // Try to send the setting in new thread.
        object : Thread() {
            override fun run() {

                try {
                    user.api.setVideoUploadsEnabled(videosEnabled)
                } catch (e: DropboxException) {
                    // Fail silently
                    DbxLog.externalError("CameraUploadsCoreControllerBase",
                            "Synching video uploads default to server failed.", e)
                }
            }
        }.start()
    }
}
