/*
 * Copyright (c) 2018, Dropbox, Inc. All rights reserved.
 */

package com.dropbox.android.camerauploads

import com.dropbox.android.user.DbxUser
import com.dropbox.android.user.DbxUserManager
import com.dropbox.android.user.UserCapability
import com.dropbox.base.error.DbxException
import com.dropbox.core.stormcrow.StormcrowBase
import com.dropbox.core.stormcrow.StormcrowMobileAndroidCameraUpload
import com.dropbox.core.stormcrow.StormcrowMobileAndroidCameraUploadAccountSwitching
import com.dropbox.core.stormcrow.StormcrowMobileAndroidCameraUploadStatusInHome

internal class CameraUploadsGatingManagerImpl(
    private val userManager: DbxUserManager
) : CameraUploadsGatingManager {
    // TODO(ageiduschek): Once account switching for biz CU is entirely enabled, delete this -
    // it will no longer be referenced
    var cuCandidateId: String? = null

    val cuEnabledUser: DbxUser?
        get() = userManager.users?.allUsers()?.filter(::isCUEnabledForUser)?.getOrNull(0)

    override fun canEnableCuForUser(user: DbxUser?): Boolean =
            user != null &&
            mayEnableCuIfCapabilityIsAvailableForUser(user) &&
            getUserCuAvailability(user) == UserCapability.Availability.AVAILABLE

    override fun shouldShowCuUIInPhotosTabForUser(user: DbxUser?): Boolean {
        if (user == null) {
            return false
        }

        return user.let {
            val hasCuEnabledForOtherUser: Boolean = cuEnabledUser?.let { it.id != user.id } ?: false
            !hasCuEnabledForOtherUser && shouldShowPhotosTabForUser(it) &&
                    (isCUEnabledForUser(it) ||
                            canShowCuEnableUIForUser(it) ||
                            shouldShowCuUpsellUIForUser(it))
        }
    }

    override fun shouldShowPhotosTabForUser(user: DbxUser?): Boolean {
        if (user == null) {
            return false
        }

        // if user is a business user and if we haven't received the Stormcrow information, we should return false
        // TODO(weilhuiz): Delete this once we've turned on
        // StormcrowMobileAndroidCameraUploadStatusInHome for 100% of business users
        try {
            val featureVariant = user.stormcrow.asStormcrowBase().currentVariants
            val stormcrowVariant =
                    featureVariant[StormcrowMobileAndroidCameraUpload.VAVAILABLE.featureName]
            if (user.role == DbxUser.UserRole.BUSINESS &&
                    StormcrowBase.VARIANT_FEATURE_NOT_RECEIVED == stormcrowVariant) {
                return false
            }
        } catch (e: DbxException) {
        }

        return !shouldShowAllCuUIInHomeForUser(user) &&
                mayEnableCuIfCapabilityIsAvailableForUser(user)
    }

    override fun shouldShowCuUpsellUIInHomeForUser(user: DbxUser?): Boolean =
            user != null &&
                    shouldShowAllCuUIInHomeForUser(user) &&
                    shouldShowCuUpsellUIForUser(user)

    override fun shouldShowCuUpsellUIForUser(user: DbxUser?): Boolean =
            user != null &&
                    mayEnableCuIfCapabilityIsAvailableForUser(user) &&
                    getUserCuAvailability(user) ==
                        UserCapability.Availability.CAN_BECOME_AVAILABLE &&
                    isPersonalOrUnpairedBusinessUser(user)

    override fun shouldShowCuEnableUIInHomeForUser(user: DbxUser?): Boolean =
            user != null &&
                    shouldShowAllCuUIInHomeForUser(user) && canShowCuEnableUIForUser(user)

    override fun shouldShowCuStatusUIInHomeForUser(user: DbxUser?): Boolean =
            user != null && shouldShowAllCuUIInHomeForUser(user) && isCUEnabledForUser(user)

    private fun shouldShowAllCuUIInHomeForUser(user: DbxUser): Boolean = try {
            user.stormcrow
                    .isInVariantLogged(
                            StormcrowMobileAndroidCameraUploadStatusInHome.VENABLED)
        } catch (e: DbxException) { false }

    private fun canShowCuEnableUIForUser(user: DbxUser): Boolean =
        mayEnableCuIfCapabilityIsAvailableForUser(user) &&
                getUserCuAvailability(user) == UserCapability.Availability.AVAILABLE &&
                !isCUEnabledForUser(user) &&
                isPersonalOrUnpairedBusinessUser(user)

    /**
     * Whether the user is allowed to enable CU if their account has the capability.
     * Before we enable account switching, a user may not be able to enable CU even if they are the
     * most capable account.
     * TODO(weilhuiz): Delete this method entirely once StormcrowMobileAndroidCameraUploadAccountSwitching
     * is enabled for all users. At that point, this method will always return true
     */
    private fun mayEnableCuIfCapabilityIsAvailableForUser(user: DbxUser): Boolean {
        val isCuAccountSwitchingEnabled = try {
            user.stormcrow
                    .isInVariantLogged(
                            StormcrowMobileAndroidCameraUploadAccountSwitching.VENABLED)
        } catch (e: DbxException) {
            false
        }

        val isCuUser = cuCandidateId == user.id

        return (isCuUser || isCuAccountSwitchingEnabled) &&
                getUserCuAvailability(user) != UserCapability.Availability.NEVER_AVAILABLE
    }

    private fun getUserCuAvailability(user: DbxUser): UserCapability.Availability =
        user.capabilitiesManager.capabilityAvailability(UserCapability.CameraUpload::class.java)

    private fun isPersonalOrUnpairedBusinessUser(user: DbxUser): Boolean =
        user.role != DbxUser.UserRole.BUSINESS || !user.isPaired

    private fun isCUEnabledForUser(user: DbxUser): Boolean =
            user.userProperties.cameraUploadEnabled()
}
