/*
 * Copyright (c) 2018, Dropbox, Inc. All rights reserved.
 */

package com.dropbox.android.fileactivity.comments

import android.support.annotation.VisibleForTesting
import com.dropbox.android.previewable.DevicePreviewableManager
import com.dropbox.android.user.DbxUser
import com.dropbox.android.user.DbxUserset
import com.dropbox.base.oxygen.DbxLog
import com.dropbox.core.DbxException
import com.dropbox.core.v2.comments2.DbxUserComments2Requests
import com.dropbox.core.v2.comments2.ErrorException
import com.dropbox.core.v2.comments2.ListThreadsResult
import com.dropbox.core.v2.comments2.Status
import com.dropbox.core.v2.comments2.Stream
import com.dropbox.core.v2.comments2.StreamIdentifier
import com.dropbox.core.v2.comments2.StreamType
import com.dropbox.hairball.path.DropboxPath
import com.dropbox.hairball.path.Path
import com.dropbox.hairball.path.SharedLinkPath
import java.util.concurrent.atomic.AtomicReference
import javax.annotation.CheckForNull
import javax.inject.Inject
import javax.inject.Singleton

/**
 * Wrapper around DbxUserComments2Requests.
 *
 * @param <P> The type of path.
 */
interface CommentsService<P : Path> {
    /**
     * Whether or not this path has associated CommentsV2 comments.
     */
    fun hasV2Comments(
        pathUser: DbxUser?,
        path: P,
        shouldCheckFileHasCommentsV2: Boolean
    ): Boolean
}

class DropboxFileCommentsService(
    private val previewableManager: DevicePreviewableManager
) : CommentsService<DropboxPath> {

    override fun hasV2Comments(
        pathUser: DbxUser?,
        path: DropboxPath,
        shouldCheckFileHasCommentsV2: Boolean
    ): Boolean {
        return checkForV2Comments(
                shouldCheckFileHasCommentsV2, previewableManager, pathUser, path)
    }
}

@Singleton
class SharedLinkCommentsService @Inject constructor(
    private val previewableManager: DevicePreviewableManager
) :
        CommentsService<SharedLinkPath> {

    var actionUser: AtomicReference<DbxUser> = AtomicReference()
        @VisibleForTesting get() = field

    override fun hasV2Comments(
        pathUser: DbxUser?,
        path: SharedLinkPath,
        shouldCheckFileHasCommentsV2: Boolean
    ): Boolean {
        val user = pathUser ?: actionUser.get()

        return checkForV2Comments(
                shouldCheckFileHasCommentsV2, previewableManager, user, path)
    }

    /**
     * When the userset changes, check if we aren't actually logged in as the action user
     * any more. If that is the case, default back to the no-auth file activity manager.
     */
    fun refreshUserset(@CheckForNull userset: DbxUserset?) {
        val actionUserRef = actionUser.get()
        if (userset == null ||
                actionUserRef != null && userset.getUserById(actionUserRef.id) == null) {
            actionUser.set(null)
        }
    }

    /**
     * Set the given user to be the preferred user for fetching/posting comments.
     */
    fun setActionUser(@CheckForNull user: DbxUser?) {
        actionUser.set(user)
    }
}

private fun <P : Path> requestCommentsStatusFromServer(
    commentsApi: DbxUserComments2Requests,
    path: P
): ListThreadsResult = commentsApi.listComments(
        Stream(StreamType.FILE, StreamIdentifier.filePathOrId(path.asCanonicalPath())))

private fun <P : Path> checkForV2Comments(
    shouldCheckFileHasCommentsV2: Boolean,
    previewableManager: DevicePreviewableManager,
    user: DbxUser?,
    path: P
): Boolean {

    if (user != null) {
        if (shouldCheckFileHasCommentsV2 &&
                previewableManager.isVideoFile(path.asCanonicalPath())) {

            val comments2 = user.apiV2.comments2()
            try {
                val result = requestCommentsStatusFromServer(comments2, path)
                if (result.status != Status.COMMENTS1) {
                    return true
                }
            } catch (e: ErrorException ) {
                DbxLog.d("CommentsService",
                        "Exception request comments status from server",
                        e)
                // if there is an exception, fall back to disable comments
                return true
            } catch (e: DbxException) {
                DbxLog.d(
                        "CommentsService",
                        "Exception request comments status from server",
                        e)
                // if there is an exception, fall back to disable comments
                return true
            }
        }
    } else {
        DbxLog.d("CommentService", "No user provided", Exception())
    }

    return false
}
