// dbx-min-coverage: 0% because it's meant to wrap untestable things
/*
 * Copyright (c) 2018, Dropbox, Inc. All rights reserved.
 */

package com.dropbox.android.payments

import android.content.ComponentName
import android.content.Context
import android.content.Intent
import android.content.ServiceConnection
import android.os.IBinder
import com.android.vending.billing.IInAppBillingService

/**
 * Interface for abstracting away the mechanics of binding to [IInAppBillingService]. Note that
 * this expects a subclass of [ServiceConnection] called [BillingServiceConnection] to ensure
 * that that only the right interfaces are provided. Implementors of this interface must be careful
 * to ensure that the service they bind to implemented [IInAppBillingService].
 */
interface BillingServiceBinder {

    /**
     * Binds the given connection to a billing service.
     */
    fun bind(connection: BillingServiceConnection): Boolean

    /**
     * Unbinds the connection.
     */
    fun unbind(connection: ServiceConnection)
}

/**
 * Specialization of [ServiceConnection] that handles service casting via final overrides of the
 * usual [ServiceConnection] methods. Subclasses want to implement the specializations instead.
 */
abstract class BillingServiceConnection : ServiceConnection {
    final override fun onServiceConnected(name: ComponentName?, service: IBinder?) {
        onBillingServiceConnected(IInAppBillingService.Stub.asInterface(service))
    }

    final override fun onServiceDisconnected(name: ComponentName?) {
        onBillingServiceDisconnected()
    }

    final override fun onBindingDied(name: ComponentName?) {
        // Don't do anything. Overridden for future proofing and to preserve the abstraction.
    }

    abstract fun onBillingServiceConnected(service: IInAppBillingService)

    abstract fun onBillingServiceDisconnected()
}

/**
 * [BillingServiceBinder] that uses the standard billing service intent and [Context.bindService].
 */
class ContextBillingServiceBinder(val context: Context) : BillingServiceBinder {
    override fun bind(connection: BillingServiceConnection): Boolean {
        val intent = Intent("com.android.vending.billing.InAppBillingService.BIND")
        intent.setPackage("com.android.vending")
        return context.bindService(intent, connection, Context.BIND_AUTO_CREATE)
    }

    override fun unbind(connection: ServiceConnection) {
        context.unbindService(connection)
    }
}
