/*
 * Copyright (c) 2018, Dropbox, Inc. All rights reserved.
 */

package com.dropbox.android.payments

import android.os.DeadObjectException
import android.os.Handler
import android.os.Looper
import android.os.RemoteException
import android.support.annotation.MainThread
import com.android.vending.billing.IInAppBillingService
import com.dropbox.android.payments.GooglePlayBindingHelper.Listener
import com.dropbox.android.util.LogTagUtils
import com.dropbox.base.oxygen.DbxLog
import com.dropbox.hairball.util.CrashLogger
import java.util.concurrent.Executor
import java.util.concurrent.RejectedExecutionException

/**
 * Helper class specifically for binding to the [IInAppBillingService].
 *
 * Handles binding and unbinding to the service and surfaces connection and disconnection events
 * via its [Listener] interface.
 */
internal class GooglePlayBindingHelper
@JvmOverloads constructor(
    val binder: BillingServiceBinder,
    val packageName: String,
    val listener: Listener,
    val handler: Handler = Handler(Looper.getMainLooper())
) {

    /**
     * Listener for callbacks after calling either [.bind] or [.unbind].
     */
    internal interface Listener {
        /**
         * Called on the main thread with the billing service when the connection to it was
         * successful and the service supports in app purchasing.
         */
        fun onConnectedAndVerified(billingService: IInAppBillingService)

        /**
         * Called on the main thread with the specific failure if connection was unsuccessful or
         * if the service does not support
         */
        fun onFailed(failure: Failure)

        /**
         * Called on the main thread when connection with the billing service is terminated. No
         * longer safe to interact with the billing service in this case. Will only be called
         * if [.onConnectedAndVerified] was previously called.
         *
         * It's possible that this won't be called after a call to [.unbind].
         */
        fun onDisconnected()
    }

    private val TAG = LogTagUtils.create(GooglePlayBindingHelper::class.java)

    // There are three stages in fetching a billing service. First a connection is established.
    // Then the connection connects and the billing service is acquired. Finally we verify that
    // in app billing is actually supported.
    private var connection: BillingServiceConnection? = null
    private var billingService: IInAppBillingService? = null
    private var billingServiceVerified: Boolean = false

    /**
     * Binds a connection to the [IInAppBillingService] using the given executor
     * to do the necessary background work. No-op if already bound.
     */
    @MainThread
    fun bind(executor: Executor) {
        if (connection != null) {
            return
        }

        val newConnection = object : BillingServiceConnection() {
            override fun onBillingServiceConnected(service: IInAppBillingService) {
                billingService = service
                billingServiceVerified = false
                try {
                    executor.execute { completeConnectionAfterVerification(billingService!!) }
                } catch (e: RejectedExecutionException) {
                    // The executor got shut down from under us, don't do anything.
                }
            }

            override fun onBillingServiceDisconnected() {
                val wasBillingServiceVerified = billingServiceVerified
                billingService = null
                billingServiceVerified = false
                if (wasBillingServiceVerified) {
                    onDisconnected()
                }
            }
        }

        var exception: Exception? = null
        var bound = false
        try {
            bound = binder.bind(newConnection)
        } catch (ex: DeadObjectException) {
            exception = ex
            // The billing service crashed was killed. Nothing we can do about this.
            DbxLog.d(TAG, "DeadObjectException trying to bind to InAppBillingService")
        } catch (ex: Exception) {
            exception = ex
            CrashLogger.getInstance().outputWarn(
                "Unknown exception when trying to bind to InAppBillingService", ex)
        }

        if (bound) {
            connection = newConnection
        } else {
            onFailed(
                Failure(
                    InAppBillingServiceUtils.BILLING_RESPONSE_RESULT_BILLING_UNAVAILABLE,
                    exception
                )
            )
        }
    }

    /**
     * Unbinds the connection to the [IInAppBillingService]. No-op if not bound.
     */
    @MainThread
    fun unbind() {
        if (connection == null) {
            return
        }

        binder.unbind(connection!!)
        connection = null
        billingService = null
        billingServiceVerified = false
    }

    private fun completeConnectionAfterVerification(billingService: IInAppBillingService) {
        val response: Int
        try {
            response = billingService.isBillingSupported(
                InAppBillingServiceUtils.GOOGLE_PLAY_API_VERSION,
                packageName,
                InAppBillingServiceUtils.ITEM_TYPE_SUBS
            )
        } catch (ex: RemoteException) {
            onFailed(Failure(InAppBillingServiceUtils.PAYMENTS_HELPER_REMOTE_EXCEPTION, ex))
            return
        }

        if (response == InAppBillingServiceUtils.BILLING_RESPONSE_RESULT_OK) {
            onConnected(billingService)
        } else {
            onFailed(Failure(response))
        }
    }

    private fun onConnected(service: IInAppBillingService) {
        handler.post {
            if (billingService === service) {
                billingServiceVerified = true
                listener.onConnectedAndVerified(service)
            }
        }
    }

    private fun onFailed(failure: Failure) {
        handler.post { listener.onFailed(failure) }
    }

    @MainThread
    private fun onDisconnected() {
        handler.post { listener.onDisconnected() }
    }
}
