/*
 * Copyright (c) 2018, Dropbox, Inc. All rights reserved.
 */

package com.dropbox.android.previewable

import com.dropbox.android.previewable.proto.PreviewableData.PreviewableExtensionInfo.PreviewType
import com.dropbox.android.user.DbxUser
import com.dropbox.core.stormcrow.NoauthStormcrow

interface DevicePreviewableManager {

    /**
     * If enough time has elapsed since the last update, this will update the dictionary used to
     * determine how files should be previewed by querying the server asynchronously. Otherwise
     * does nothing.
     */
    fun updateInfoAsyncIfNeeded()

    /**
     * @return the preferred ways to preview the contents of a file. The returned list is ordered by
     * preference (i.e. first element in the list is the most preferable).
     */
    fun getPreferredPreviewTypes(fileName: String): List<PreviewType>

    /**
     * Checks to see if the given file type supports video streaming from Riviera.
     * Examples include: .mov, .mpg.
     * @param fileName The name or path of the file.
     * @return true if the file can be streamed
     */
    fun isStreamingVideoFile(fileName: String): Boolean

    /**
     * Checks to see if a file is a video file. Note that this is different from a video stream.
     * Examples include: .mp4, .mov, .qt.
     * @param fileName The name or path of the file.
     * @return true if the file is a video file
     */
    fun isVideoFile(fileName: String): Boolean

    /**
     * Checks to see if a file is a web link file. Web link files are simply shortcuts to a webpage.
     * Examples include: .website, .url, .webloc.
     * @param fileName The name or path of the file.
     * @return true if the file is a web link file
     */
    fun isWebLink(fileName: String): Boolean

    /**
     * Checks to see if a file is a text based file. Examples include: .txt, .cpp, .bat
     * @param fileName The name or path of the file.
     * @return true if the file is a text based file
     */
    fun isTextFile(fileName: String): Boolean

    /**
     * Checks to see if a file is a transcoded video file. A transcoded video file is a video file
     * that the server supports previews for and thus has transcoded the video into another,
     * streamable format.
     * @param fileName The name or path of the file.
     * @return true if the file is a text based file
     */
    fun isTranscodableVideoFile(fileName: String): Boolean

    /**
     * Checks to see if a file has an image preview. Note that this does not necessarily mean that
     * the file is actually an image, for instance a .sketch file may have an image preview even
     * though .sketch files are actually complicated file types.
     * streamable format.
     * @param fileName The name or path of the file.
     * @return true if the file has an image preview.
     */
    fun hasImagePreview(fileName: String): Boolean

    /**
     * Returns whether the provided user has permission to preview the given [PreviewType]. If
     * no user is given, we will check against the [NoauthStormcrow] instead.
     */
    fun shouldShowPreviewForType(previewType: PreviewType, user: DbxUser?): Boolean
}
