/*
 * Copyright (c) 2017, Dropbox, Inc. All rights reserved.
 */

package com.dropbox.android.previewable

import android.content.Context
import com.dropbox.android.BuildConfig
import com.dropbox.android.R
import com.dropbox.android.previewable.proto.PreviewableData
import com.dropbox.android.previewable.proto.PreviewableData.PreviewableExtensionInfo
import com.dropbox.android.previewable.proto.PreviewableData.PreviewableExtensionInfo.PreviewType
import com.dropbox.android.settings.GlobalProperties
import com.dropbox.android.user.DbxUser
import com.dropbox.android.util.ApiNetworkException
import com.dropbox.hairball.util.FileNameUtils
import com.dropbox.base.inject.ApplicationContext
import com.dropbox.base.inject.LateInit
import com.dropbox.base.json.*
import com.dropbox.base.oxygen.DbxAssert
import com.dropbox.base.thread.NamedThreadFactory
import com.dropbox.core.DbxException
import com.dropbox.core.stormcrow.*
import com.google.common.annotations.VisibleForTesting
import org.joda.time.Duration
import org.joda.time.Instant
import org.json.simple.parser.JSONParser
import org.json.simple.parser.ParseException
import java.io.IOException
import java.io.InputStreamReader
import java.util.*
import java.util.concurrent.Executor
import java.util.concurrent.Executors
import java.util.concurrent.atomic.AtomicBoolean
import java.util.concurrent.atomic.AtomicLong
import java.util.concurrent.atomic.AtomicReference
import javax.inject.Inject
import javax.inject.Singleton

private val REFRESH_INTERVAL = Duration.standardDays(1L)

/**
 * Class that manages information regarding which types of files are previewable and how they can be
 * previewed.
 *
 * Note that file types here are determined by the extension of the file and not the MIME type.
 */
@Singleton
class DevicePreviewableManagerImpl
    @VisibleForTesting
    internal constructor(
        context: Context,
        private val properties: GlobalProperties,
        private val stormcrow: LateInit<NoauthStormcrow>,
        private val previewableApi: PreviewableApi,
        private val executor: Executor
    ) : DevicePreviewableManager {

    private val context: Context = context.applicationContext

    private val initialized: AtomicBoolean = AtomicBoolean()
    private val updating: AtomicBoolean = AtomicBoolean()

    /**
     * A ref to a map from file extensions to [PreviewableExtensionInfo].
     */
    private val previewableDictionaryRef: AtomicReference<PreviewableDictionary> =
            AtomicReference(PreviewableDictionary())
    private val lastUpdateMs: AtomicLong = AtomicLong(0)

    @Inject
    internal constructor(
        @ApplicationContext context: Context,
        properties: GlobalProperties,
        stormcrow: LateInit<NoauthStormcrow>,
        previewableApi: PreviewableApi
    ) : this(
            context,
            properties,
            stormcrow,
            previewableApi,
            Executors.newSingleThreadExecutor(
                    NamedThreadFactory.from(DevicePreviewableManager::class.java)
                            .useDefaultFactory()))

    /**
     * Initializes this manager. Must be called before using.
     *
     * Note that this method needs to be synchronized otherwise a race condition can occur where the
     * data has not been loaded yet but initialize() will return.
     */
    private fun initialize() {
        synchronized(this) {
            if (initialized.compareAndSet(false, true)) {
                loadData(properties.previewableData)
            }
        }
    }

    /**
     * Loads the previewable information from a protobuf object.
     */
    private fun loadData(previewableData: PreviewableData) {
        val timeDataRetrieved = previewableData.timeRetrieved

        if (timeDataRetrieved == 0L) {
            // We don't actually have any cached previewable data. Use the offline data instead
            val offlineData = getOfflineData(context)

            if (offlineData != null) {
                loadData(offlineData)
                return
            }
        }

        loadData(previewableData.previewableInfoList)

        val dataAppVersion = if (previewableData.hasAppVersionCode())
            previewableData.appVersionCode else 0

        if (dataAppVersion < BuildConfig.VERSION_CODE) {
            // we always want to refresh if the user just updated their app
            lastUpdateMs.set(0)
        } else if (previewableData.hasTimeRetrieved()) {
            lastUpdateMs.set(timeDataRetrieved)
        }
    }

    /**
     * Loads the previewable information from a map.
     */
    private fun loadData(previewableInfo: Collection<PreviewableExtensionInfo>) {
        val previewableDictionary = PreviewableDictionary()

        // Note that the order we add items to the PreviewableDictionary matters. The items added
        // first have higher priority than the ones added later.

        // We will only add the first item in the FullScreenPreviewTypesList and ignore the rest.
        // This was a design decision. The reason for this is as follows:
        // 1) Server currently only supports having 1 full content preview type per file type.
        // 2) Server wants to be able to support more than 1 type in the future.
        // 3) Server wants to send back a list for full content preview types due to 2). However at
        //    the time of writing this list will contain at most 1 element due to 1).
        // 4) The server might return more than 1 full content preview type in the future. To
        //    ensure that the client doesn't break in the future we will only look at the first
        //    item in the full content preview type. This is because the current previews endpoint
        //    (the on used at the time of writing) doesn't take in a full content preview type as an
        //    argument and thus cannot support multiple preview types.
        // Note that we have safeguard on the server in the form of unit tests that ensure that the
        // server will not break the client in the future.
        for (info in previewableInfo) {
            if (info.fullScreenPreviewTypesList.isNotEmpty()) {
                previewableDictionary.add(
                        info.fileExtension,
                        listOf(info.fullScreenPreviewTypesList[0]))
            }
        }

        addHardcodedExtensions(previewableDictionary)

        previewableDictionaryRef.set(previewableDictionary)
    }

    /**
     * Retrieves updated information from the API endpoint if enough time has passed since the last update.
     *
     * @return `true` if API information was refreshed successfully
     */
    @VisibleForTesting
    fun internalUpdateInfo(): Boolean {
        DbxAssert.notMainThread()

        initialize()

        if (updating.getAndSet(true)) {
            return false
        }

        val previewableExtensions: Map<String, PreviewableExtensionInfo>
        try {
            previewableExtensions = previewableApi.previewableExtensions
        } catch (e: ApiNetworkException) {
            return false
        } catch (e: DbxException) {
            return false
        }

        loadData(previewableExtensions.values)
        saveData(previewableExtensions.values)
        lastUpdateMs.set(System.currentTimeMillis())

        updating.set(false)

        return true
    }

    /**
     * Serializes the in-memory presentations of the user data and saves the protobuf to user properties.
     */
    private fun saveData(previewableInfo: Collection<PreviewableExtensionInfo>) {
        val previewableData = PreviewableData.newBuilder()
                .addAllPreviewableInfo(previewableInfo)
                .setTimeRetrieved(System.currentTimeMillis())
                .setAppVersionCode(BuildConfig.VERSION_CODE)
                .build()

        properties.previewableData = previewableData
    }

    /**
     * If enough time has elapsed since the last update, this will update the dictionary used to
     * determine how files should be previewed by querying the server asynchronously. Otherwise
     * does nothing.
     */
    override fun updateInfoAsyncIfNeeded() {
        val needUpdate = { (Instant.now() - REFRESH_INTERVAL).millis >= lastUpdateMs.get() }

        if (initialized.get() && needUpdate()) {
            return
        }

        executor.execute {
            initialize()

            if (needUpdate()) {
                internalUpdateInfo()
            }
        }
    }

    /**
     * Force refreshes this manager. Should only be used for testing purposes.
     */
    @VisibleForTesting
    fun forceUpdateInfo() {
        DbxAssert.notMainThread()

        initialize()
        internalUpdateInfo()
    }

    /**
     * @return the preferred way to preview the contents of a file.  The returned list is ordered by
     * preference (i.e. first element in the list is the most preferable).
     */
    override fun getPreferredPreviewTypes(fileName: String): List<PreviewType> {
        initialize()

        val ext = FileNameUtils.getEndingLoweredWithoutPeriod(fileName)
        return previewableDictionaryRef.get().getPreviewTypesFor(ext)
    }

    /**
     * Checks to see if the given file type supports video streaming from Riviera.
     * Examples include: .mov, .mpg.
     * @param fileName The name or path of the file.
     * @return true if the file can be streamed
     */
    override fun isStreamingVideoFile(fileName: String): Boolean {
        return getPreferredPreviewTypes(fileName).contains(PreviewType.STREAMING_VIDEO)
    }

    /**
     * Checks to see if a file is a video file. Note that this is different from a video stream.
     * Examples include: .mp4, .mov, .qt.
     * @param fileName The name or path of the file.
     * @return true if the file is a video file
     */
    override fun isVideoFile(fileName: String): Boolean {
        return getPreferredPreviewTypes(fileName).contains(PreviewType.VIDEO)
    }

    /**
     * Checks to see if a file is a web link file. Web link files are simply shortcuts to a webpage.
     * Examples include: .website, .url, .webloc.
     * @param fileName The name or path of the file.
     * @return true if the file is a web link file
     */
    override fun isWebLink(fileName: String): Boolean {
        return getPreferredPreviewTypes(fileName).contains(PreviewType.LINK_FILE)
    }

    /**
     * Checks to see if a file is a text based file. Examples include: .txt, .cpp, .bat
     * @param fileName The name or path of the file.
     * @return true if the file is a text based file
     */
    override fun isTextFile(fileName: String): Boolean {
        return getPreferredPreviewTypes(fileName).contains(PreviewType.TEXT)
    }

    /**
     * Checks to see if a file is a transcoded video file. A transcoded video file is a video file
     * that the server supports previews for and thus has transcoded the video into another,
     * streamable format.
     * @param fileName The name or path of the file.
     * @return true if the file is a text based file
     */
    override fun isTranscodableVideoFile(fileName: String): Boolean {
        return getPreferredPreviewTypes(fileName).contains(PreviewType.STREAMING_VIDEO)
    }

    /**
     * Checks to see if a file has an image preview. Note that this does not necessarily mean that
     * the file is actually an image, for instance a .sketch file may have an image preview even
     * though .sketch files are actually complicated file types.
     * streamable format.
     * @param fileName The name or path of the file.
     * @return true if the file has an image preview.
     */
    override fun hasImagePreview(fileName: String): Boolean {
        return getPreferredPreviewTypes(fileName).any {
            it == PreviewType.THUMBNAIL || it == PreviewType.STREAMING_VIDEO
        }
    }

    /**
     * Returns whether the provided user has permission to preview the given [PreviewType]. If
     * no user is given, we will check against the [NoauthStormcrow] instead.
     */
    override fun shouldShowPreviewForType(
        previewType: PreviewType,
        user: DbxUser?
    ): Boolean {
        try {
            if (previewType == PreviewType.HTML) {
                return user?.stormcrow?.isInVariantUnlogged(
                        StormcrowMobileAndroidDocPreviewsHtml.VSHOW_HTML_PREVIEWS)
                        ?: stormcrow
                                .get()
                                .isInNoauthVariantUnlogged(
                                        StormcrowMobileAndroidDocPreviewsHtmlNouser
                                                .VSHOW_HTML_PREVIEWS)
            } else if (previewType == PreviewType.PDF || previewType == PreviewType.PDF_PREVIEW) {
                return user?.stormcrow
                        ?.isInVariantUnlogged(StormcrowMobileAndroidDocPreviews.VDOC_PREVIEWS)
                        ?: stormcrow
                                .get()
                                .isInNoauthVariantUnlogged(
                                        StormcrowMobileAndroidDocPreviewsNouser.VDOC_PREVIEWS)
            }
        } catch (e: com.dropbox.base.error.DbxException) {
            // Just return a sane default.
            return false
        }

        // We default to show all preview types unless we want to explicitly disallow previews of
        // the file type.
        return true
    }

    private class PreviewableDictionary {
        private val dictionary = mutableMapOf<String, MutableList<PreviewType>>()

        fun add(extension: String, toAdd: List<PreviewType>) {
            val previewTypes = dictionary.getOrPut(extension, { mutableListOf<PreviewType>() })

            previewTypes.addAll(toAdd.filter {
                !previewTypes.contains(it)
            })
        }

        /**
         * Convenience method to add a single preview type to multiple extensions.
         */
        fun addPreviewTypeForExtensions(previewType: PreviewType, extensions: List<String>) {
            for (extension in extensions) {
                add(extension, listOf(previewType))
            }
        }

        fun getPreviewTypesFor(extension: String): List<PreviewType> {
            val previewTypes = dictionary[extension]
            return previewTypes ?: listOf()
        }
    }

    companion object {
        val PREVIEWABLE_EXTENSIONS_RESULT_JSON_EXTRACTOR =
                object : JsonExtractor<List<PreviewableExtensionInfo>>() {

            override fun extract(jt: JsonThing): List<PreviewableExtensionInfo> {
                val data = jt.expectMap()
                val fileExtensionMap = data.get("file_extension_to_preview_info").expectMap()

                return fileExtensionMap.map { (ext, value) ->
                    extractPreviewInfo(ext, value.expectMap())
                }
            }

            private fun extractPreviewInfo(ext: String, map: JsonMap): PreviewableExtensionInfo {
                val previewTypesJson = map.get("preview_types").expectList()
                val fullScreenPreviewTypesJson = map.get("full_screen_preview_types").expectList()

                val builder = PreviewableExtensionInfo.newBuilder()

                val previewTypes = extractPreviewTypes(previewTypesJson)
                val fullScreenPreviewTypes = extractPreviewTypes(fullScreenPreviewTypesJson)

                return builder.setFileExtension(ext)
                        .addAllFullScreenPreviewTypes(fullScreenPreviewTypes)
                        .addAllSupportedPreviewTypes(previewTypes)
                        .build()
            }

            private fun extractPreviewTypes(previewTypesJson: JsonList): List<PreviewType> {
                val previewTypes = previewTypesJson.map {
                    com.dropbox.core.v2.files.PreviewType.valueOf(
                            it.expectMap().get(".tag").expectString().toUpperCase(Locale.US))
                }
                return PreviewableApiConversions.previewTypesSdkToClient(previewTypes)
            }
        }

        /**
         * Returns the preview information that can be accessed offline. This is also useful as a
         * backup in case the server preview info has not been retrieved yet or if the server preview
         * info is unable to be retrieved.
         */
        @VisibleForTesting
        @JvmStatic
        fun getOfflineData(context: Context): List<PreviewableExtensionInfo>? {
            val parser = JSONParser()

            val inputStream = context.resources.openRawResource(R.raw.previewableinfo)

            try {
                val jsonObject = parser.parse(InputStreamReader(inputStream))
                val jsonThing = JsonThing(jsonObject)

                return PREVIEWABLE_EXTENSIONS_RESULT_JSON_EXTRACTOR.extract(jsonThing)
            } catch (e: IOException) {
                return null
            } catch (e: ParseException) {
                return null
            } catch (e: JsonExtractionException) {
                return null
            }
        }

        /**
         * This is a hard coded list of extensions that we know how to preview the originals of.
         * This is hard coded because it can be platform specific and since Riviera only returns
         * previews and not originals.
         */
        private fun addHardcodedExtensions(previewableDictionary: PreviewableDictionary) {
            previewableDictionary.addPreviewTypeForExtensions(
                    PreviewType.TEXT,
                    listOf(
                            "abap",
                            "ada",
                            "adp",
                            "ahk",
                            "as",
                            "as3",
                            "asc",
                            "ascx",
                            "asm",
                            "asp",
                            "aspx",
                            "awk",
                            "bash",
                            "bash_login",
                            "bash_logout",
                            "bash_profile",
                            "bashrc",
                            "bat",
                            "bib",
                            "bsh",
                            "build",
                            "builder",
                            "c",
                            "c++",
                            "capfile",
                            "cc",
                            "cfc",
                            "cfm",
                            "cfml",
                            "cl",
                            "clj",
                            "cls",
                            "cmake",
                            "cmd",
                            "coffee",
                            "cpp",
                            "cpt",
                            "cpy",
                            "cs",
                            "cshtml",
                            "cson",
                            "csproj",
                            "css",
                            "ctp",
                            "cxx",
                            "d",
                            "ddl",
                            "di",
                            "dif",
                            "diff",
                            "disco",
                            "dml",
                            "dtd",
                            "dtml",
                            "el",
                            "emakefile",
                            "erb",
                            "erl",
                            "f",
                            "f90",
                            "f95",
                            "fs",
                            "fsi",
                            "fsscript",
                            "fsx",
                            "gemfile",
                            "gemspec",
                            "gitconfig",
                            "go",
                            "groovy",
                            "gry",
                            "gvy",
                            "h",
                            "h++",
                            "haml",
                            "handlebars",
                            "hbs",
                            "hcp",
                            "hh",
                            "hpp",
                            "hrl",
                            "hs",
                            "htc",
                            "htm",
                            "html",
                            "hxx",
                            "idl",
                            "iim",
                            "inc",
                            "inf",
                            "ini",
                            "inl",
                            "ipp",
                            "irbrc",
                            "jade",
                            "jav",
                            "java",
                            "js",
                            "json",
                            "jsp",
                            "jsx",
                            "l",
                            "less",
                            "lhs",
                            "lisp",
                            "log",
                            "lst",
                            "ltx",
                            "lua",
                            "m",
                            "make",
                            "markdn",
                            "markdown",
                            "md",
                            "mdown",
                            "mkdn",
                            "ml",
                            "mli",
                            "mll",
                            "mly",
                            "mm",
                            "mud",
                            "nfo",
                            "opml",
                            "osascript",
                            "out",
                            "p",
                            "pas",
                            "patch",
                            "php",
                            "php2",
                            "php3",
                            "php4",
                            "php5",
                            "phtml",
                            "pl",
                            "plist",
                            "pm",
                            "pod",
                            "pp",
                            "profile",
                            "properties",
                            "ps1",
                            "pt",
                            "py",
                            "pyw",
                            "r",
                            "rake",
                            "rb",
                            "rbx",
                            "rc",
                            "re",
                            "readme",
                            "reg",
                            "rest",
                            "resw",
                            "resx",
                            "rhtml",
                            "rjs",
                            "rprofile",
                            "rpy",
                            "rss",
                            "rst",
                            "rxml",
                            "s",
                            "sass",
                            "scala",
                            "scm",
                            "sconscript",
                            "sconstruct",
                            "script",
                            "scss",
                            "sgml",
                            "sh",
                            "shtml",
                            "sml",
                            "sql",
                            "sty",
                            "tcl",
                            "tex",
                            "text",
                            "textile",
                            "tld",
                            "tli",
                            "tmpl",
                            "tpl",
                            "txt",
                            "vb",
                            "vi",
                            "vim",
                            "wsdl",
                            "xhtml",
                            "xml",
                            "xoml",
                            "xsd",
                            "xsl",
                            "xslt",
                            "yaml",
                            "yaws",
                            "yml",
                            "zsh",
                            "zshabap"))

            previewableDictionary.addPreviewTypeForExtensions(
                    PreviewType.LINK_FILE, listOf("url", "website", "webloc"))

            previewableDictionary.addPreviewTypeForExtensions(
                    PreviewType.AUDIO,
                    listOf(
                            "3ga", "aa", "aac", "aax", "aif", "aifc", "aiff", "amr", "au", "flac",
                            "kar", "m3u", "m4a", "m4b", "m4r", "mid", "midi", "mka", "mp2", "mp3",
                            "mpga", "oga", "ra", "ram", "rm", "snd", "wav", "wma"))

            previewableDictionary.addPreviewTypeForExtensions(
                    PreviewType.VIDEO,
                    listOf(
                            "3gp", "3gpp", "avi", "mkv", "mov", "movie", "mp4", "mpe", "mpeg",
                            "mpg", "mxu", "qt", "rmvb"))

            previewableDictionary.addPreviewTypeForExtensions(PreviewType.PDF, listOf("pdf"))
        }
    }
}
