package com.dropbox.android.settings

import com.dropbox.android.settings.DBPreferences.DBEditor
import com.dropbox.android.util.Protobuf
import com.dropbox.base.oxygen.DbxLog
import com.google.protobuf.MessageLite
import com.google.protobuf.Parser

/**
 * A SettingsTransaction is for atomically grouping a set of Setting changes. All
 * actions must match the SettingsTransaction's initially specified source. SettingsTransaction
 * operations are only applied if commit() is called.
 */
class SettingsTransaction(private val source: PrefsSource) {
    private val editor: DBEditor = source.getPrefs().edit()

    fun getEditor(source: PrefsSource): DBEditor {
        if (this.source !== source) {
            throw RuntimeException("Scope mismatch!")
        }
        return editor
    }

    fun commit(): Boolean {
        return editor.commit()
    }
}

interface PrefsSource {
    fun getPrefs(): DBPreferences
}

abstract class BaseSetting<T : Any> protected constructor(
    protected val prefsSource: PrefsSource,
    protected val key: String
) {

    /**
     * Returns true if any value exists for this Setting.
     */
    val isSet: Boolean
        get() = prefs.contains(key)

    /**
     * Returns the appropriate DBPreferences object for this Setting's
     * specified Scope.
     */
    protected val prefs: DBPreferences
        get() = prefsSource.getPrefs()

    init {
        if (RESERVED_KEYS.contains(key)) {
            throw AssertionError(key + " is a reserved key and should not be used")
        }
    }

    /**
     * remove value of mKey on disk, execute on a new editor so operations in editor
     * without commit won't be committed
     */
    fun removeAndCommit() {
        prefs.edit().remove(key).commit()
    }

    /**
     *  remove value of mKey in editor without commit yet
     */
    fun remove(transaction: SettingsTransaction) {
        transaction.getEditor(prefsSource).remove(key)
    }

    /**
     * get value of key that is already commit
     */
    abstract fun get(): T?

    /**
     * put value of a key without commit yet
     */
    abstract fun put(transaction: SettingsTransaction, value: T)

    /**
     * put value of a key and commit
     */
    abstract fun putAndCommit(value: T)

    companion object {

        /**
         * Keys are reserved if they used to be valid, so they may still have
         * data on some devices. This just keeps us from accidentally reusing
         * them.
         */
        private val RESERVED_KEYS = setOf(
                "IN_CU_SNOOZED_EXPERIMENT",
                "CAMERA_UPLOADS_ALBUM_REV",
                "LAST_RESYNC_TS",
                "PHOTO_SELECT_POPUP_DISMISSED_NEW",
                "BROMOS_DISMISSED",
                "BROMOS_METADATA",
                "SHARING_FLOW_STAGE",
                "CAMERA_UPLOAD_BROMO_MAX_IDS",
                "CAMERA_UPLOAD_BROMO_COUNT",
                "CAMERA_UPLOAD_SEEN_BROMO",
                "SEEN_TOUR",
                "DID_REPORT_HOST_SPECIAL",
                "PAIRING_INFO",
                "FAVORITES_MIGRATED_TO_DATASTORE",
                "FAVORITES_SYNC_ENABLED",
                "CAROUSEL_CROSSOVER_CONFIG",
                "ANAL_NEXT_ROTATION")
    }
}

class BoolSetting(
    prefsSource: PrefsSource,
    key: String,
    private val default: Boolean
) : BaseSetting<Boolean>(prefsSource, key) {

    override fun get(): Boolean {
        return prefs.getBoolean(key, default)
    }

    override fun putAndCommit(value: Boolean) {
        prefs.edit().putBoolean(key, value).commit()
    }

    override fun put(transaction: SettingsTransaction, value: Boolean) {
        transaction.getEditor(prefsSource).putBoolean(key, value)
    }
}

class LongSetting(
    prefsSource: PrefsSource,
    key: String,
    private val default: Long
) : BaseSetting<Long>(prefsSource, key) {

    override fun get(): Long {
        return prefs.getLong(key, default)
    }

    override fun putAndCommit(value: Long) {
        prefs.edit().putLong(key, value).commit()
    }

    override fun put(transaction: SettingsTransaction, value: Long) {
        transaction.getEditor(prefsSource).putLong(key, value)
    }
}

class IntSetting(
    prefsSource: PrefsSource,
    key: String,
    private val default: Int
) : BaseSetting<Int>(prefsSource, key) {

    override fun get(): Int {
        return prefs.getInt(key, default)
    }

    override fun putAndCommit(value: Int) {
        prefs.edit().putInt(key, value).commit()
    }

    override fun put(transaction: SettingsTransaction, value: Int) {
        transaction.getEditor(prefsSource).putInt(key, value)
    }
}

class FloatSetting(
    prefsSource: PrefsSource,
    key: String,
    private val default: Float
) : BaseSetting<Float>(prefsSource, key) {

    override fun get(): Float {
        return prefs.getFloat(key, default)
    }

    override fun putAndCommit(value: Float) {
        prefs.edit().putFloat(key, value).commit()
    }

    override fun put(transaction: SettingsTransaction, value: Float) {
        transaction.getEditor(prefsSource).putFloat(key, value)
    }
}

class StringSetting(
    prefsSource: PrefsSource,
    key: String,
    private val default: String?
) : BaseSetting<String>(prefsSource, key) {

    override fun get(): String? {
        return prefs.getString(key, default)
    }

    override fun putAndCommit(value: String) {
        prefs.edit().putString(key, value).commit()
    }

    override fun put(transaction: SettingsTransaction, value: String) {
        transaction.getEditor(prefsSource).putString(key, value)
    }
}

class EnumSetting<T : Enum<T>>(
    prefsSource: PrefsSource,
    key: String,
    private val myClass: Class<T>,
    private val default: T?
) : BaseSetting<T>(prefsSource, key) {

    override fun get(): T? {
        val value = prefs.getString(key, null) ?: return default
        return try {
            java.lang.Enum.valueOf<T>(myClass, value)
        } catch (e: IllegalArgumentException) {
            DbxLog.d("EnumSetting",
                    "Ignoring exception and returning default value " +
                            "after failed enum cast.",
                    e)
            default
        }
    }

    override fun putAndCommit(value: T) {
        prefs.edit().putString(key, value.name).commit()
    }

    override fun put(transaction: SettingsTransaction, value: T) {
        transaction.getEditor(prefsSource).putString(key, value.name)
    }
}

class MessageSetting<T : MessageLite>(
    prefsSource: PrefsSource,
    key: String,
    private val myParser: Parser<T>,
    private val default: T?
) : BaseSetting<T>(prefsSource, key) {

    constructor(prefsSource: PrefsSource, key: String, def: T) :
            this(prefsSource, key, Protobuf.getParserForType<T>(def), def)

    override fun get(): T? {
        val base64 = prefs.getString(key, null)
        return if (base64 != null) Protobuf.decode<T>(base64, myParser) else default
    }

    override fun putAndCommit(value: T) {
        prefs.edit().putString(key, Protobuf.encode(value)).commit()
    }

    override fun put(transaction: SettingsTransaction, value: T) {
        transaction.getEditor(prefsSource).putString(key, Protobuf.encode(value))
    }
}
